/*
 Phone Clusters.js
 Copyright 2012 Greg J. Hedlund

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */

/*
params = 
	{separator, "<html><b>Information</b></html>", true},
	{label, "<html><p>Phone Clusters.js (version 3)</p></html>", "<html><b>Script:</b></html>"},
	{label, "<html><p>Greg J. Hedlund <a href='mailto:ghedlund@mun.ca'>ghedlund@mun.ca</a></p></html>", "<html><b>Author:</b></html>"},
	{label, "<html><p>24 April 2012</p></html>", "<html><b>Created:</b></html>"},
	{label, "<html><p>Apache Licence, Version 2 <a href='http://www.apache.org/licenses/LICENSE-2.0'>http://www.apache.org/licenses/LICENSE-2.0</a><html>", "<html><b>Licence:</b></html>"},
	{label, "<html><p>1.5.2</p></html>", "<html><b>Phon Version:</b></html>"},
	{label, "<html><p>Search for consonant or vowel clusters within words.  Words can be selected by position within the group as well as stress pattern.</p></html>", "<html><b>Description:</b></html>"},
	{label, "", "<html><b>Changes</b></html>"},
	{label, "", "<html>30-April-2012:</html>"},
	{label, "<html><p>Added support for clusters separated by syllable boundaries.</p></html>", ""},
	{label, "<html><p>Added pre-defined cluster filters.</p></html>", ""},
	{label, "", "<html>01-May-2012:</html>"},
	{label, "<html><p>Fixed a bug when adding aligned phones to the result.</p></html>", ""},
	{separator, "<html><b>Cluster Type</b></html>", false},
		{enum, tierName, "IPA Target"|"IPA Actual", 0, "<html><b>Tier:</b></html>"},
		{enum, clusterType, "Consonant"|"Vowel", 0, "<html><b>Type:</b></html>"},
		{label, "", "<html><b>Filter:</b></html>"},
		{enum, filterSelection, "Anything"|"Singleton"|"Tuple"|"Heterosyllabic consonants"|"Other (specifiy below)",
			0,
			"<html>Pre-defined:</html>"},
		{string, clusterPhonex, "", "<html>Phonex:</html>"},	
		{label, "<html><p>{} = singleton; {}{} = sequence of two consonants; {}+ = one or more of anything; etc.</p></html>", "<html>Key:</html>"},
		{label, "<html><p>{} can be specified for features (e.g. {Coronal}), syllable positions (e.g. {}:Coda {}:Onset)</p></html>", "<html>Note:</html>"},
		{label, "<html><p>{}:OEHS|Coda|RightAppendix+ {}:SyllableBoundaryMarker|SyllableStressMarker? {}:OEHS|LeftAppendix|Onset+</p></html>", "<html>Heterosyllabic consonant clusters:</html>"},
		{multibool,
			cInitial|cMedial|cFinal,
			true|true|true,
			"Initial"|"Medial"|"Final",
			"<html><b>Cluster Position in Word:</b></html>", 3},
	{separator, "<html><b>Word Position & Stress Pattern</b></html>", false},
		{bool, wSingleton, true, "(groups with only one word)", "Singleton words:"},
		{multibool,
			wInitial|wMedial|wFinal,
			true|true|true,
			"Initial"|"Medial"|"Final",
			"Multiple words:", 3},
		{string, wordStressPattern, "A+", "<html><b>Stress Pattern:</b></html>"},
		{label, "<html><b>1</b> = primary stress, <b>2</b> = secondary stress, <b>U</b> = no stress</html>", "Key:"},
		{label, "<html><b>A</b> = anything</html>", "Special:"},
		{label, "<html><b>*</b> = zero or more, <b>+</b> = one or more, <b>?</b> = zero or one</html>", "Quantifiers:"},
	{separator, "<html><b>Result Options</b></html>", false},
		{bool, includeAligned, true, "<html>Include aligned phones with result</html>", "<html><b>Aligned Phones:</b></html>"},
		{label, "", "<html><b>Metadata:</b></html>"},
		{bool, includeStressPattern, true, "<html>Include word stress pattern (e.g., 1U2)</html>", ""}
	;
 */

/********************************
 * Define globals
 *******************************/
var session;

var CONSONANT_CLUSTER_PHONEX = "{C}+";
var VOWEL_CLUSTER_PHONEX = "{V}+";

var predefinedExprs = [ 
	"{}+", 
	"{}", 
	"{}{}", 
	"{}:OEHS|Coda|RightAppendix+ {}:SyllableBoundaryMarker|SyllableStressMarker? {}:OEHS|LeftAppendix|Onset+"
];

/********************************
 * query_record (required)
 *
 * Called for each record in a session.
 * Perform search operations here.
 *
 * params:
 *	record - current record
 * returns:
 *	void
 *******************************/
function query_record(record)
{
	var phonex = 
		(clusterType.index == 0 ? CONSONANT_CLUSTER_PHONEX : VOWEL_CLUSTER_PHONEX);

	var filter = 
		(filterSelection.index < 4 ? predefinedExprs[filterSelection.index] : clusterPhonex);
	
	for(i = 0; i < record.numberOfGroups; i++)
	{
		var ipaGrp = record.getGroup(tierName, i);

		var ipaWords = getRequestedWords(ipaGrp);
		
		for(j = 0; j < ipaWords.length; j++)
		{
			var ipaWord = ipaWords[j];
			// convert results to a javascript array
			// for the convienent 'splice' method
			var clusters = new Array();
			var tempClusters = ipaWord.findPhonex(phonex);
			for(k = 0; k < tempClusters.length; k++) clusters.push(tempClusters[k]);
			var medialClustersWithStress = 
				ipaWord.findPhonex(phonex + "{}:SyllableBoundaryMarker|SyllableStressMarker"+phonex);

			for(k = 0; k < medialClustersWithStress.length; k++) {
				var spliceStart = -1;
				var numToRemove = 0;

				var c1 = medialClustersWithStress[k];
				for(l = 0; l < clusters.length; l++) {
					var c2 = clusters[l];

					if(c1.dataRange.contains(c2.dataRange)) {
						spliceStart = 
							(spliceStart == -1 ? l : spliceStart);
						numToRemove++;
					}
				}
				
				if(spliceStart >= 0) {
					clusters.splice(spliceStart, numToRemove, c1);
				}
			}
			
			for(k = 0; k < clusters.length; k++) 
			{
				var cluster = clusters[k];
				var clusterOk = false;
				if(filter.length > 0 && cluster.matchesPhonex(filter))
					clusterOk = true;

				var posOk = true;
				if(is_word_initial(ipaWord, cluster) && !cInitial) posOk = false;
				if(is_word_medial(ipaWord, cluster) && !cMedial) posOk = false;
				if(is_word_final(ipaWord, cluster) && !cFinal) posOk = false;
				
				if(clusterOk & posOk) 
				{
					var result = create_result(record, ipaWord, cluster);
					results.add(result.value, result.metadata, result.type);
				}
			}
		}
	}
}

/********************************
 * begin_search (optional)
 *
 * Called once at the beginning of each
 * search.  Initialize globals and perform
 * any other startup tasks.
 *
 * params:
 *	s - the session being searched
 * returns:
 *	void
 *******************************/
 function begin_search(s)
 {
 	// store session in a global variable
 	session = s;
 	
 	/*
 	 * Loading a Custom Feature set:
 	 *
 	 * To load a custom feature set for this
 	 * search un-comment the following lines
 	 * and replace <file> with a properly
 	 * formatted features file.
 	 * For more information see 
 	 * http://phon.ling.mun.ca/phontrac/wiki/search/customfeatures
 	 */
	// var customFM = new FeatureMatrix("<file>");
	// FeatureMatrix.makeCurrentInThread( customFM );
 }
 
 /********************************
 * end_search (optional)
 *
 * 
 * Perform any necessary cleanup operations.
 * Use this function to report data to the
 * console at the end of a search.
 *
 * params:
 *	s - the session being searched
 * returns:
 *	void
 *******************************/
function end_search(s)
{
	// do cleanup
	
	/*
	 * If a custom feature matrix has been
	 * loaded un-comment the following line
	 * to reset to default FeatureMatrix
	 * for future searches.
	 */
	// FeatureMatrix.makeCurrentInThread( null );
}

/**
 * Is the given cluster word initial
 * 
 * @param word the parent word
 * @param cluster the cluster in question
 * @returns <code>true</code> if the given cluster
 *  is word initial, <code>false</code> false otherwise
 */
function is_word_initial(word, cluster)
{
	var retVal = false;
	// check index of first phone in cluster
	// if it's 0 - it's word initial. 
	// if it's 1 - check the first phone in the word, if
	//  it's a stress marker the cluster is also word initial
	var firstClusterPhone = cluster.getPhone(0);

	var phoneOffset = word.getPhone(0).phone.phoneIndex;
	var clusterPhoneIndex = 
		firstClusterPhone.phone.phoneIndex - phoneOffset;

	if(clusterPhoneIndex == 0) {
		retVal = true;
	} else if(clusterPhoneIndex == 1) {
		if(word.getPhone(0).scType.identifier == "SS") {
			retVal = true;
		}
	}
	
	return retVal;
}

/**
 * Is the given cluster word final
 * 
 * @param word the parent word
 * @param cluster the cluster in question
 * @returns <code>true</code> if the given cluster
 *  is word final, <code>false</code> false otherwise
 */
function is_word_final(word, cluster)
{
	var retVal = false;
	
	var lastClusterPhone = cluster.getPhone(cluster.numberOfPhones-1);
	var lastWordPhone = word.getPhone(word.numberOfPhones-1);
	
	if(lastClusterPhone.phone.phoneIndex == lastWordPhone.phone.phoneIndex) {
		retVal = true;
	}
	
	return retVal;
}

function is_word_medial(word, cluster) 
{
	var retVal = !is_word_initial(word, cluster) && !is_word_final(word, cluster);
	return retVal;
}

/**
 * Create result
 * 
 * @param record
 * @param word the parent word
 * @param cluster the cluster to add
 * @returns [result, type] the created result along
 *  with the result 'schema' type
 */
function create_result(record, word, cluster)
{
	var retVal = [ cluster ];
	var resultType = "LINEAR";
	
	if(includeAligned) 
	{
		var aligned = record.getAlignedPhones(cluster);
		retVal[1] = aligned;
		resultType = "ALIGNED";
	}

	return { 
		value: retVal, 
		type: resultType, 
		metadata: create_metadata(word, cluster) 
	};
}

/**
 * Create metadata for result
 * 
 * @param word the parent word
 * @param cluster the cluster
 * @returns the result metadata
 */
function create_metadata(word, cluster)
{
	var retVal = new Metadata();

	if(includeStressPattern) {
		var pattern = create_stress_pattern(word);
		retVal.put("Stress Pattern", pattern);
	}

	return retVal;
}

/**
 * Create a text stress pattern
 * for the given word
 * 
 * @param word
 * @returns the word stress pattern
 */
function create_stress_pattern(word)
{
	var retVal = "";

	for(var sIdx = 0; sIdx < word.numberOfSyllables; sIdx++)
	{
		var syllable = word.getSyllable(sIdx);

		switch(syllable.stress)
		{
			case 0:
			retVal += "U";
			break;
			
			case 1:
			retVal += "1";
			break;

			case 2:
			retVal += "2";
			break;

			default:
			break;
		}
	}

	return retVal;
}

/**
 * Returns a list of words for the given
 * group which match the criteria in the form.
 */
function getRequestedWords(group)
{
	var retVal = new Array();
	var retIdx = 0;

	if(wSingleton && group.numberOfWords == 1) {
		var word = group.getWord(0);
		if(wordStressPattern.length > 0 && word.matchesStressPattern(wordStressPattern))
			retVal[retIdx++] = group.getWord(0);
	} else {
		for(var wIndex = 0; wIndex < group.numberOfWords; wIndex++)
		{
			var word = group.getWord(wIndex);

			var posOk = true;
			if(wIndex == 0 && !wInitial) posOk = false;
			if(wIndex > 0 && wIndex < group.numberOfWords-1 && !wMedial) posOk = false;
			if(wIndex == group.numberOfWords-1 && !wFinal) posOk = false;

			var patternOk = true;
			if(wordStressPattern.length > 0)
				patternOk = word.matchesStressPattern(wordStressPattern);
				
			if(posOk && patternOk)
			{
				retVal[retIdx++] = word;
			}
		}
	}

	return retVal;
}
